﻿using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Problems.DataAnalysis.Symbolic;

namespace HeuristicLab.Encodings.SymbolicExpressionTreeEncoding {
  public class SymbolicExpressionTreeNodeComparer : IEqualityComparer<ISymbolicExpressionTreeNode> {

    public SymbolicExpressionTreeNodeComparer(int similarityLevel) {
      _similarityLevel = similarityLevel;
    }

    public int GetHashCode(ISymbolicExpressionTreeNode n) {
      return n.ToString().ToLower().GetHashCode();
    }

    public bool Equals(ISymbolicExpressionTreeNode a, ISymbolicExpressionTreeNode b) {
      if (a.SubtreeCount != b.SubtreeCount) { return false; }
      if (a.SubtreeCount > 0) { return a.Symbol.ToString().Equals(b.Symbol.ToString()); }
      // compare leaf nodes according to desired similarity level
      switch (_similarityLevel) {
        case ((int)SymbolicExpressionTreeMatching.SimilarityLevel.Exact):
          if (a is ConstantTreeNode && b is ConstantTreeNode) {
            return ((ConstantTreeNode)a).Value.Equals(((ConstantTreeNode)b).Value);
          }
          if (a is VariableTreeNode && b is VariableTreeNode) {
            return (a as VariableTreeNode).Weight.Equals((b as VariableTreeNode).Weight);
          }
          return false;
        case ((int)SymbolicExpressionTreeMatching.SimilarityLevel.High):
          return (a is ConstantTreeNode && b is ConstantTreeNode) || (a is VariableTreeNode && b is VariableTreeNode);
        case ((int)SymbolicExpressionTreeMatching.SimilarityLevel.Relaxed):
          return true;
        default:
          return false;
      }
    }

    private readonly int _similarityLevel;
  }

  // tree equality
  public class SymbolicExpressionTreeComparer : IEqualityComparer<ISymbolicExpressionTree> {
    public bool Equals(ISymbolicExpressionTree a, ISymbolicExpressionTree b) {
      return SymbolicExpressionTreeMatching.FindMatch(a, b, _mode) == 0;
    }

    public int GetHashCode(ISymbolicExpressionTree tree) {
      return tree.Length;
    }

    private int _mode;
    public void SetComparisonMode(int mode) {
      _mode = mode;
    }
  }

  public static class SymbolicExpressionTreeMatching {
    public enum SimilarityLevel {
      Exact,   // everything is matched bit by bit (functions and terminals)
      High,    // symbols are matched. leaf node types are matched
      Relaxed  // only symbols are matched, leafs count as wildcards
    }

    public static bool IsSimilarTo(this ISymbolicExpressionTree t1, ISymbolicExpressionTree t2, int mode) {
      return FindMatch(t1, t2, mode) == 0;
    }
    public static bool IsSimilarTo(this ISymbolicExpressionTreeNode t1, ISymbolicExpressionTreeNode t2, int mode) {
      return FindMatch(t1, t2, mode) == 0;
    }

    public static bool ContainsFragment(this ISymbolicExpressionTree tree, ISymbolicExpressionTreeNode fragment, int mode) {
      var nodes = tree.IterateNodesPostfix().ToArray();
      var fragments = fragment.IterateNodesPostfix().ToArray();
      return FindMatch(nodes, fragments, mode) != -1;
    }

    #region FindMatch signatures
    public static int FindMatch(ISymbolicExpressionTree t0, ISymbolicExpressionTree t1, int mode) {
      return FindMatch(t0.IterateNodesPostfix().ToArray(), t1.IterateNodesPostfix().ToArray(), mode);
    }
    private static int FindMatch(ISymbolicExpressionTreeNode t0, ISymbolicExpressionTreeNode t1, int mode) {
      return FindMatch(t0.IterateNodesPostfix().ToArray(), t1.IterateNodesPostfix().ToArray(), mode);
    }
    private static int FindMatch(ISymbolicExpressionTree t0, ISymbolicExpressionTreeNode t1, int mode) {
      return FindMatch(t0.IterateNodesPostfix().ToArray(), t1.IterateNodesPostfix().ToArray(), mode);
    }
    private static int FindMatch(ISymbolicExpressionTreeNode t0, ISymbolicExpressionTree t1, int mode) {
      return FindMatch(t0.IterateNodesPostfix().ToArray(), t1.IterateNodesPostfix().ToArray(), mode);
    }
    #endregion

    // convenience methods for less typing :)
    private static IEnumerator<ISymbolicExpressionTreeNode> Enumerator(this ISymbolicExpressionTree tree) {
      return tree.IterateNodes().GetEnumerator();
    }
    private static IEnumerator<ISymbolicExpressionTreeNode> Enumerator(this ISymbolicExpressionTreeNode tree) {
      return tree.IterateNodes().GetEnumerator();
    }
    // method for breath-width iteration of nodes
    public static IEnumerable<ISymbolicExpressionTreeNode> IterateNodes(this ISymbolicExpressionTree tree) {
      return IterateNodes(tree.Root);
    }

    private static IEnumerable<ISymbolicExpressionTreeNode> IterateNodes(this ISymbolicExpressionTreeNode root) {
      var list = new List<ISymbolicExpressionTreeNode> { root };
      int offset = 0, count = 1;
      while (offset != count) {
        var c = count;
        for (int i = offset; i != count; ++i) {
          yield return list[i]; // TODO: look into yield performance (vs returning the whole list at the end)
          if (!list[i].Subtrees.Any()) continue;
          list.AddRange(list[i].Subtrees);
        }
        offset = c;
        count = list.Count;
      }
      //return list;
    }

    /// <summary>
    /// This method finds the point where parent and child differ. For instance in the case of crossover, 
    /// it will return the node that was swapped into parent0 from parent1
    /// Prerequisites: as the naming implies, heredity should exist between parent and child,
    /// otherwise this method would not make much sense (it would always return the root of child)
    /// </summary>
    /// <param name="parent">The individual from the previous generation that was crossed over or mutated to produce child</param>
    /// <param name="child">The result of a genetic operation applied on parent</param>
    /// <returns>A symbolic expression tree node representing the difference fragment between parent and child</returns>
    public static ISymbolicExpressionTreeNode GetFragmentDiff(ISymbolicExpressionTree parent, ISymbolicExpressionTree child) {
      var nodes1 = parent.IterateNodes().ToList();
      var nodes2 = child.IterateNodes().ToList();
      var e1 = nodes1.AsEnumerable().GetEnumerator();
      var e2 = nodes2.AsEnumerable().GetEnumerator();
      while (e1.MoveNext() && e2.MoveNext()) {
        var comparer = new SymbolicExpressionTreeNodeComparer((int)SimilarityLevel.Exact);
        if (!comparer.Equals(e1.Current, e2.Current)) {
          return e2.Current;
        }
        //if (!e1.Current.IterateNodes().SequenceEqual(e2.Current.IterateNodes(), comparer)) {
        //  return e2.Current;
        //} // return the fragment by which child differs from parent
      }
      return null;
    }

    // returns the index in the sequence 'seq', where pattern 'pat' is matched, or -1 for no match
    // -1 means no match, 0 means perfect match (the sequences overlap, hence the match index is 0)
    public static int FindMatch(ISymbolicExpressionTreeNode[] seq, ISymbolicExpressionTreeNode[] pat, int mode) {
      int patlen = pat.Length;
      int seqlen = seq.Length;
      if (patlen == 0 || seqlen == 0) return -1;
      var comparer = new SymbolicExpressionTreeNodeComparer(mode);
      if (patlen == seqlen) return seq.SequenceEqual(pat, comparer) ? 0 : -1;
      for (int i = patlen; i <= seqlen; ++i) {
        if (comparer.Equals(seq[i - 1], pat.Last())) // first check if last element is a match
          if (seq.Skip(i - patlen).Take(patlen).SequenceEqual(pat, comparer)) // then check the whole sequence
            return i - patlen;
      }
      return -1;
    }
  }
}
